/*
 * Normaliz 2.5
 * Copyright (C) 2007-2010  Winfried Bruns, Bogdan Ichim, Christof Soeger
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

/**
 * The class Sublattice_Representation represents a sublattice of Z^n as Z^r.
 * To transform vectors of the sublattice  use:
 *    Z^r --> Z^n    and    Z^n -->  Z^r
 *     v  |-> vA             u  |-> (uB)/c
 * A  r x n matrix
 * B  n x r matrix
 * c  Integer
 */

#ifndef SUBLATTICE_REPRESENTATION_H
#define SUBLATTICE_REPRESENTATION_H

#include "lineare_transformation.h"

//---------------------------------------------------------------------------

class Sublattice_Representation {
	int dim, rank;
	Matrix A;
	Matrix B;
	Integer c;
	Integer index;

//---------------------------------------------------------------------------
public:
//---------------------------------------------------------------------------
//						Construction and destruction
//---------------------------------------------------------------------------

	/**
	 * creates a dummy object
	 */
	Sublattice_Representation() {};

	/**
	 * creates a representation of Z^n as a sublattice of itself
	 */
	Sublattice_Representation(int n);
	
	/**
	 * Main Constructor
	 * creates a representation of a sublattice of Z^n
	 * if direct_summand is false the sublattice is generated by the rows of M
	 * otherwise it is a direct summand of Z^n containing the rows of M
	 */
	Sublattice_Representation(const Matrix& M, bool direct_summand);
	Sublattice_Representation(const Lineare_Transformation& LT, bool direct_summand);
	Sublattice_Representation(const Sublattice_Representation& SR);
	~Sublattice_Representation();


//---------------------------------------------------------------------------
//                       Manipulation operations
//---------------------------------------------------------------------------	

	/* like the matching constructor */
	void initialize(const Lineare_Transformation& LT, bool direct_summand);

	/* first this then SR when going from Z^n to Z^r */
	void compose(const Sublattice_Representation& SR);

//---------------------------------------------------------------------------
//                       Transformations
//---------------------------------------------------------------------------

	Matrix to_sublattice (const Matrix& M) const;
	Matrix from_sublattice (const Matrix& M) const;
	Matrix to_sublattice_dual (const Matrix& M) const;
	Matrix from_sublattice_dual (const Matrix& M) const;

	vector<Integer> to_sublattice (const vector<Integer>& V) const;
	vector<Integer> from_sublattice (const vector<Integer>& V) const;
	vector<Integer> to_sublattice_dual (const vector<Integer>& M) const;
	vector<Integer> from_sublattice_dual (const vector<Integer>& V) const;

//---------------------------------------------------------------------------
//						 Data acces
//---------------------------------------------------------------------------

	/* returns the dimension of the ambient space */
	int get_dim() const;

	/* returns the rank of the sublattice */
	int get_rank() const;

	/* returns the index of the sublattice */
	Integer get_index() const;

	Matrix get_A() const;
	
	Matrix get_B() const;
	
	Integer get_c() const;
	
	/* returns the congruences defining the sublattice */
	Matrix get_congruences() const;

};
//---------------------------------------------------------------------------
#endif
//---------------------------------------------------------------------------
